# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************

from typing import Dict, Callable, Tuple, List

from ..core import utils
from .csv_logging import CSVLogger
from .common_stream import CommonStream
from ..core.enums.dcb_enums import DCBCommand
from ..core.enums.ecg_enums import ECGCommand
from ..core.packets.command_packet import CommandPacket
from ..core.packets.stream_data_packets import ECGDataPacket
from ..core.enums.common_enums import Application, Stream, CommonCommand
from ..core.packets.common_packets import DecimationFactorPacket, DCBPacket, VersionPacket, LibraryConfigReadWritePacket


class ECGApplication(CommonStream):
    """
    ECG Application class.

    .. code-block:: python3
        :emphasize-lines: 4

        from adi_study_watch import SDK

        sdk = SDK("COM4")
        application = sdk.get_ecg_application()

    """

    def __init__(self, callback_function, packet_manager, args):
        super().__init__(Application.ECG, Stream.ECG, packet_manager, callback_function, args)
        self._dcb_size = 2

    def delete_device_configuration_block(self) -> Dict:
        """
        Deletes ECG Device configuration block.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ecg_application()
            application.delete_device_configuration_block()
        """
        packet = DCBPacket(self._destination, DCBCommand.ERASE_CONFIG_REQ)
        packet.set_dcb_data_size(0)
        return self._send_packet(packet, DCBCommand.ERASE_CONFIG_RES)

    def get_decimation_factor(self) -> Dict:
        """
        Returns stream decimation factor.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ecg_application()
            x = application.get_decimation_factor()
            print(x["payload"]["decimation_factor"])
            # 1
        """
        packet = DecimationFactorPacket(self._destination, CommonCommand.GET_STREAM_DEC_FACTOR_REQ)
        packet.set_stream_address(self._stream)
        return self._send_packet(packet, CommonCommand.GET_STREAM_DEC_FACTOR_RES)

    def get_version(self) -> Dict:
        """
        Returns ECG version info.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ecg_application()
            x = application.get_version()
            print(x["payload"]["major_version"])
            # 3
            print(x["payload"]["minor_version"])
            # 4
            print(x["payload"]["patch_version"])
            # 3
            print(x["payload"]["version_string"])
            # ECG_App
            print(x["payload"]["build_version"])
            # TEST ECG_VERSION STRING
        """
        packet = VersionPacket(self._destination, CommonCommand.GET_VERSION_REQ)
        return self._send_packet(packet, CommonCommand.GET_VERSION_RES)

    def get_algo_version(self) -> Dict:
        """
        Returns ECG version info.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ecg_application()
            x = application.get_algo_version()
            print(x["payload"]["major_version"])
            # 3
            print(x["payload"]["minor_version"])
            # 4
            print(x["payload"]["patch_version"])
            # 3
            print(x["payload"]["version_string"])
            # ECG_App
            print(x["payload"]["build_version"])
            # TEST ECG_VERSION STRING
        """
        packet = VersionPacket(self._destination, ECGCommand.GET_ALGO_VENDOR_VERSION_REQ)
        return self._send_packet(packet, ECGCommand.GET_ALGO_VENDOR_VERSION_RES)

    def read_device_configuration_block(self) -> Dict:
        """
        Returns entire device configuration block.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ecg_application()
            x = application.get_ecg_application()
            print(x["payload"]["dcb_data"])
            # []

        """
        packet = DCBPacket(self._destination, DCBCommand.READ_CONFIG_REQ)
        packet.set_dcb_data_size(0)
        return self._send_packet(packet, DCBCommand.READ_CONFIG_RES)

    def read_library_configuration(self, fields: List[int]) -> Dict:
        """
        Reads library configuration from specified field values.

        :param fields: List of field values to read.
        :type fields: List[int]
        :return: A response packet as dictionary
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Fields Lower Limit
             - Fields Upper Limit
           * - 0x00
             - 0x02

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ecg_application()
            x = application.read_library_configuration([0x00])
            print(x["payload"]["data"])
            # [['0x0', '0x0']]
        """
        address_range = [0x00, 0x02]
        packet = LibraryConfigReadWritePacket(self._destination, CommonCommand.READ_LCFG_REQ)
        packet.set_number_of_operations(len(fields))
        utils.check_array_address_range(fields, address_range, 1)
        packet.set_read_fields_data(fields)
        return self._send_packet(packet, CommonCommand.READ_LCFG_RES)

    def set_decimation_factor(self, decimation_factor: int) -> Dict:
        """
        Sets decimation factor for ECG stream.

        :param decimation_factor: decimation factor for stream
        :type decimation_factor: int
        :return: A response packet as dictionary
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Decimation Lower Limit
             - Decimation Upper Limit
           * - 0x01
             - 0x05

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ecg_application()
            x = application.set_decimation_factor(2)
            print(x["payload"]["decimation_factor"])
            # 2
        """
        decimation_factor = utils.range_and_type_check(decimation_factor, type_of=int, num_bytes=1)
        packet = DecimationFactorPacket(self._destination, CommonCommand.SET_STREAM_DEC_FACTOR_REQ)
        packet.set_stream_address(self._stream)
        packet.set_decimation_factor([decimation_factor])
        return self._send_packet(packet, CommonCommand.SET_STREAM_DEC_FACTOR_RES)

    def write_device_configuration_block(self, addresses_values: List[List[int]]) -> Dict:
        """
        Writes the device configuration block values of specified addresses.
        This function takes a list of addresses and values to write, and returns a response packet as
        dictionary containing addresses and values.

        :param addresses_values: List of addresses and values to write.
        :type addresses_values: List[List[int]]
        :return: A response packet as dictionary.
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Address Lower Limit
             - Address Upper Limit
           * - 0x00
             - 0x02

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ecg_application()
            x = application.write_device_configuration_block([[0x0, 2], [0x1, 0x1]])
            print(x["payload"]["num_ops"])
            # 2
        """
        address_range = [0x00, 0x02]
        packet = DCBPacket(self._destination, DCBCommand.WRITE_CONFIG_REQ)
        num_ops = len(addresses_values)
        packet.set_size(utils.split_int_in_bytes(num_ops, length=2))
        utils.check_array_address_range(addresses_values, address_range, 2)
        packet.set_dcb_write_data(addresses_values, self._dcb_size, num_ops)
        return self._send_packet(packet, DCBCommand.WRITE_CONFIG_RES)

    def write_device_configuration_block_from_file(self, filename: str) -> Dict:
        """
        Writes the device configuration block values of specified addresses from file.

        :param filename: dcb filename
        :return: A response packet as dictionary.
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Address Lower Limit
             - Address Upper Limit
           * - 0x00
             - 0x02

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ecg_application()
            application.write_device_configuration_block_from_file("ecg_dcb.lcfg")
        """
        result = self._write_device_configuration_block_from_file_helper(filename)
        if result:
            return self.write_device_configuration_block(result)

    def write_dcb_to_lcfg(self) -> Dict:
        """
        Writes Device configuration block data to library configuration.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ecg_application()
            x = application.write_dcb_to_lcfg()
            print(x["payload"]["status"])
            # CommonStatus.OK
        """
        packet = CommandPacket(self._destination, CommonCommand.SET_LCFG_REQ)
        return self._send_packet(packet, CommonCommand.SET_LCFG_RES)

    def write_library_configuration(self, fields_values: List[List[int]]) -> Dict:
        """
        Writes library configuration from List of fields and values.

        :param fields_values: List of fields and values to write.
        :type fields_values: List[List[int]]
        :return: A response packet as dictionary.
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Fields Lower Limit
             - Fields Upper Limit
           * - 0x00
             - 0x03

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ecg_application()
            x = application.write_library_configuration([[0x01, 0x2]])
            print(x["payload"]["data"])
            # [['0x0', '0x1']]

        """
        address_range = [0x00, 0x03]
        packet = LibraryConfigReadWritePacket(self._destination, CommonCommand.WRITE_LCFG_REQ)
        packet.set_number_of_operations(len(fields_values))
        utils.check_array_address_range(fields_values, address_range, 2)
        packet.set_write_fields_data(fields_values)
        return self._send_packet(packet, CommonCommand.WRITE_LCFG_RES)

    def get_sensor_status(self) -> Dict:
        """
        Returns packet with number of subscribers and number of sensor start request registered.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ecg_application()
            x = application.get_sensor_status()
            print(x["payload"]["num_subscribers"], x["payload"]["num_start_registered"])
            # 0 0
        """
        return super().get_sensor_status()

    def set_callback(self, callback_function: Callable, args: Tuple = ()) -> None:
        """
        Sets the callback for the stream data.

        :param args: optional arguments that will be passed with the callback.
        :param callback_function: callback function for stream ECG data.
        :return: None

        .. code-block:: python3
            :emphasize-lines: 4,12

            from adi_study_watch import SDK

            # make sure optional arguments have default value to prevent them causing Exceptions.
            def callback(data, optional1=None, optional2=None):
                print(data)

            sdk = SDK("COM4")
            application = sdk.get_ecg_application()
            # these optional arguments can be used to pass file, matplotlib or other objects to manipulate data.
            optional_arg1 = "1"
            optional_arg2 = "2"
            application.set_callback(callback, args=(optional_arg1, optional_arg2))
        """
        super().set_callback(callback_function, args)

    def set_timeout(self, timeout_value: float) -> None:
        """
        Sets the time out for queue to wait for command packet response.

        :param timeout_value: queue timeout value (in sec).
        :type timeout_value: int
        :return: None

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ecg_application()
            application.set_timeout(10)
        """
        super().set_timeout(timeout_value)

    def start_and_subscribe_stream(self) -> Tuple[Dict, Dict]:
        """
        Starts ECG sensor and also subscribe to the ECG stream.

        :return: A response packet as dictionary.
        :rtype: Tuple[Dict, Dict]

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ecg_application()
            start_sensor, subs_stream = application.start_and_subscribe_stream()
            print(start_sensor["payload"]["status"], subs_stream["payload"]["status"])
            # CommonStatus.STREAM_STARTED CommonStatus.SUBSCRIBER_ADDED
        """
        return super().start_and_subscribe_stream()

    def start_sensor(self) -> Dict:
        """
        Starts ECG sensor.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ecg_application()
            start_sensor = application.start_sensor()
            print(start_sensor["payload"]["status"])
            # CommonStatus.STREAM_STARTED
        """
        return super().start_sensor()

    def stop_and_unsubscribe_stream(self) -> Tuple[Dict, Dict]:
        """
        Stops ECG sensor and also Unsubscribe the ECG stream.

        :return: A response packet as dictionary.
        :rtype: Tuple[Dict, Dict]

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ecg_application()
            stop_sensor, unsubscribe_stream = application.stop_and_unsubscribe_stream()
            print(stop_sensor["payload"]["status"], unsubscribe_stream["payload"]["status"])
            # CommonStatus.STREAM_STOPPED CommonStatus.SUBSCRIBER_REMOVED
        """
        return super().stop_and_unsubscribe_stream()

    def stop_sensor(self) -> Dict:
        """
        Stops ECG sensor.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ecg_application()
            stop_sensor = application.stop_sensor()
            print(stop_sensor["payload"]["status"])
            # CommonStatus.STREAM_STOPPED
        """
        return super().stop_sensor()

    def subscribe_stream(self) -> Dict:
        """
        Subscribe to the ECG stream.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ecg_application()
            subs_stream = application.subscribe_stream()
            print(subs_stream["payload"]["status"])
            # CommonStatus.SUBSCRIBER_ADDED
        """
        return super().subscribe_stream()

    def unsubscribe_stream(self) -> Dict:
        """
        Unsubscribe the ECG stream.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ecg_application()
            unsubscribe_stream = application.unsubscribe_stream()
            print(unsubscribe_stream["payload"]["status"])
            # CommonStatus.SUBSCRIBER_REMOVED
        """
        return super().unsubscribe_stream()

    def _callback_data(self, packet, packet_id, callback_function=None, args=None):
        """
        Process and returns the data back to user's callback function.
        """
        self._callback_data_helper(packet, packet_id, ECGDataPacket(), callback_function, args)

    def enable_csv_logging(self, filename, header=None) -> None:
        """
        Start logging stream data into CSV.

        :param filename: Name of the CSV file.
        :param header: Header list of the CSV file.
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ecg_application()
            x = application.enable_csv_logging("ecg.csv")
        """
        if header is None:
            header = ["Timestamp", "Seq No.", "ECG data"]
        self._csv_logger[Stream.ECG] = CSVLogger(filename, header)

    def disable_csv_logging(self) -> None:
        """
        Stops logging stream data into CSV.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ecg_application()
            x = application.disable_csv_logging()
        """
        if self._csv_logger[Stream.ECG]:
            self._csv_logger[Stream.ECG].stop_logging()
        self._csv_logger[Stream.ECG] = None



# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************

from .. import utils
from .command_packet import CommandPacket
from ..enums.bcm_enums import BCMDFTWindow
from ..enums.eda_enums import EDADFTWindow
from ..enums.common_enums import Stream, Application
from ..enums.adpd_enums import ADPDCommand, ADPDSlot, ADPDAppID


class DCBPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.AD7156: ['0xC8', '0x0B']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x5C',
                'checksum': '0x0'
            },
            'payload': {
                'command': <DCBCommand.READ_CONFIG_RES: ['0x98']>,
                'status': <DCBStatus.OK: ['0x97']>,
                'size': 2,
                'dcb_data': [
                    [ '0x1F', '0xFF' ],
                    [ '0x0', '0x7' ]
                ]
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["size"] = {"size": 2, "join": True}
        self._config["payload"]["dcb_data"] = {"size": 8}
        self._packet["payload"]["size"] = [0x00, 0x00]
        self._packet["payload"]["dcb_data"] = [0x00] * 8

    def set_size(self, size):
        self._packet["payload"]["size"] = size

    def set_dcb_data_size(self, size):
        self._packet["payload"]["dcb_data"] = [0x00, 0x00, 0x00, 0x00] * size

    def set_dcb_write_data(self, dcb_data, dcb_size, num_ops):
        dcb_data_array = []
        if self._packet["header"]["destination"] == Application.PPG.value:
            dcb_data_array = utils.split_int_in_bytes(0, length=dcb_size * 4)
            for i in range(num_ops):
                start_index = i * 4
                dcb_data_array[start_index: start_index + 4] = utils.split_int_in_bytes(dcb_data[i], length=4)
        elif self._packet["header"]["destination"] in [Application.ADXL.value, Application.AD7156.value,
                                                       Application.PM.value]:
            dcb_data_array = utils.split_int_in_bytes(0, length=dcb_size * 4)
            for i in range(num_ops):
                start_index = i * 4
                dcb_data_array[start_index] = dcb_data[i][1]
                dcb_data_array[start_index + 1] = dcb_data[i][0]
        elif self._packet["header"]["destination"] == Application.ADPD.value:
            dcb_data_array = utils.split_int_in_bytes(0, length=(dcb_size * 4))
            for i in range(num_ops):
                start_index = i * 4
                dcb_data_array[start_index: start_index + 2] = utils.split_int_in_bytes(dcb_data[i][1], length=2)
                dcb_data_array[start_index + 2: start_index + 4] = utils.split_int_in_bytes(dcb_data[i][0], length=2)
        elif self._packet["header"]["destination"] == Application.LT_APP.value:
            dcb_data_array = dcb_data
        elif self._packet["header"]["destination"] in [Application.ECG.value, Application.EDA.value,
                                                       Application.BCM.value]:
            dcb_data_array = utils.split_int_in_bytes(0, length=dcb_size * 4)
            for i in range(num_ops):
                start_index = i * 4
                dcb_data_array[start_index: start_index + 2] = utils.split_int_in_bytes(dcb_data[i][1], length=2)
                dcb_data_array[start_index + 2] = dcb_data[i][0]
        self._packet["payload"]["dcb_data"] = dcb_data_array

    def decode_packet(self, data):
        super().decode_packet(data)
        dcb_data = data[12:]
        size = self._packet["payload"]["size"]
        result = []
        if self._packet["header"]["source"] == Application.PPG:
            for i in range(size):
                start_index = i * 4
                dcb_value = utils.join_multi_length_packets(dcb_data[start_index: start_index + 4], convert_to_hex=True)
                result.append([i, dcb_value])
        elif self._packet["header"]["source"] in [Application.ADXL, Application.AD7156, Application.PM]:
            for i in range(size):
                i = i * 4
                dcb_value = utils.join_multi_length_packets(dcb_data[i: i + 1], convert_to_hex=True)
                dcb_address = utils.join_multi_length_packets(dcb_data[i + 1: i + 2], convert_to_hex=True)
                result.append([dcb_address, dcb_value])
        elif self._packet["header"]["source"] == Application.ADPD:
            self._packet["payload"]["num_of_packets"] = utils.join_multi_length_packets(data[12:14])
            dcb_data = data[14:]
            for i in range(size):
                i = i * 4
                dcb_value = utils.join_multi_length_packets(dcb_data[i: i + 2], convert_to_hex=True)
                dcb_address = utils.join_multi_length_packets(dcb_data[i + 2: i + 4], convert_to_hex=True)
                result.append([dcb_address, dcb_value])
        elif self._packet["header"]["source"] == Application.LT_APP:
            self._packet["payload"]["num_of_packets"] = utils.join_multi_length_packets(data[12:14])
            result = data[14:]
        elif self._packet["header"]["source"] in [Application.ECG, Application.EDA, Application.BCM]:
            for i in range(size):
                i = i * 4
                dcb_value = utils.join_multi_length_packets(dcb_data[i: i + 2], convert_to_hex=True)
                dcb_address = utils.join_multi_length_packets(dcb_data[i + 2: i + 3], convert_to_hex=True)
                result.append([dcb_address, dcb_value])

        self._packet["payload"]["dcb_data"] = result


class DCFGPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.ADPD: ['0xC1', '0x10']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0xF0',
                'checksum': '0x0'
            },
            'payload': {
                'command': <CommonCommand.GET_DCFG_RES: ['0x26']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'size': 57,
                'num_tx_packets': 1,
                'data': [
                    [ '0x9', '0x9C' ],
                    [ '0x7', '0x8FFF' ],
                    [ '0xB', '0x322' ],
                            ...
                    [ '0x1B5', '0x0' ],
                    [ '0x1B6', '0x0' ]
                ]
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["size"] = {"size": 1, "join": True}
        self._config["payload"]["num_tx_packets"] = {"size": 1, "join": True}
        self._config["payload"]["dcfg_data"] = {"size": -1}

    def decode_packet(self, data):
        super().decode_packet(data)
        address_value_data = self._packet["payload"]["dcfg_data"]
        num_ops = self._packet["payload"]["size"]
        data = []
        if self._packet["header"]["source"] in [Application.TEMPERATURE, Application.ADPD]:
            for i in range(num_ops):
                start_index = i * 4
                value_data = address_value_data[start_index:start_index + 2]
                value = utils.join_multi_length_packets(value_data, convert_to_hex=True)
                address_data = address_value_data[start_index + 2:start_index + 4]
                address = utils.join_multi_length_packets(address_data, convert_to_hex=True)
                data.append([address, value])
        else:
            for i in range(num_ops):
                start_index = i * 2
                value_data = address_value_data[start_index:start_index + 1]
                value = utils.join_multi_length_packets(value_data, convert_to_hex=True)
                address_data = address_value_data[start_index + 1:start_index + 2]
                address = utils.join_multi_length_packets(address_data, convert_to_hex=True)
                data.append([address, value])
        self._packet["payload"]["data"] = data
        del self._packet["payload"]["dcfg_data"]


class DecimationFactorPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.ADPD: ['0xC1', '0x10']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0xD',
                'checksum': '0x0'
            },
            'payload': {
                'command': <CommonCommand.GET_STREAM_DEC_FACTOR_RES: ['0x2A']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'stream_address': <Stream.ADPD6: ['0xC2', '0x16']>,
                'decimation_factor': 1
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["stream_address"] = {"size": 2}
        self._config["payload"]["decimation_factor"] = {"size": 1, "join": True}
        self._packet["payload"]["stream_address"] = [0x00, 0x00]
        self._packet["payload"]["decimation_factor"] = [0x00]

    def set_decimation_factor(self, decimation_factor):
        self._packet["payload"]["decimation_factor"] = decimation_factor

    def set_stream_address(self, stream):
        self._packet["payload"]["stream_address"] = list(reversed(stream.value))

    def decode_packet(self, data):
        super().decode_packet(data)
        stream_address = self._packet["payload"]["stream_address"]
        self._packet["payload"]["stream_address"] = Stream(list(reversed(stream_address)))


class LibraryConfigDataPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.PPG: ['0xC3', '0x00']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0xDF',
                'checksum': '0x0'
            },
            'payload': {
                'command': <CommonCommand.GET_LCFG_RES: ['0x13']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'size': 53,
                'lcfg_data': [
                    [ 0, '0x0' ],
                    [ 1, '0x0' ],
                        ...
                    [ 52, '0x0' ]
                ]
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["size"] = {"size": 1, "join": True}
        self._config["payload"]["lcfg_data"] = {"size": -1}
        self._packet["payload"]["size"] = [0x00]
        self._packet["payload"]["lcfg_data"] = [0x00] * (4 * 53)

    def decode_packet(self, data):
        super().decode_packet(data)
        size = self._packet["payload"]["size"]
        lcfg_data = self._packet["payload"]["lcfg_data"]
        data = []
        for i in range(size):
            start_index = i * 4
            value = utils.join_multi_length_packets(lcfg_data[start_index:start_index + 4], convert_to_hex=True)
            data.append([i, value])
        self._packet["payload"]["lcfg_data"] = data


class SetLibraryConfigPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.PPG: ['0xC3', '0x00']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0xA',
                'checksum': '0x0'
            },
            'payload': {
                'command': <CommonCommand.SET_LCFG_RES: ['0x15']>,
                'status': <CommonStatus.OK: ['0x00']>
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)

    def set_lcfg_id(self, lcfg_id):
        self._packet["payload"]["lcfg_id"] = lcfg_id


class LibraryConfigReadWritePacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.ADPD: ['0xC1', '0x10']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0xE',
                'checksum': '0x0'
            },
            'payload': {
                'command': <CommonCommand.READ_LCFG_RES: ['0x17']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'num_ops': 1,
                'data': [
                    [ '0x0', '0x12C' ]
                ]
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["num_ops"] = {"size": 1, "join": True}
        self._packet["payload"]["num_ops"] = [0x01]
        self._packet["payload"]["data"] = [0x00, 0x00, 0x00, 0x00, 0x00]

    def set_number_of_operations(self, num_ops):
        self._packet["payload"]["num_ops"] = [num_ops]

    def set_write_fields_data(self, data):
        num_ops = len(data)
        if self._packet["header"]["destination"] in [Application.ECG.value, Application.ADPD.value,
                                                     Application.LT_APP.value]:
            fields_array = utils.split_int_in_bytes(0, length=num_ops * 3)
            for i in range(num_ops):
                start_index = i * 3
                fields_array[start_index] = data[i][0]
                fields_array[start_index + 1: start_index + 3] = utils.split_int_in_bytes(data[i][1], length=2)
            self._packet["payload"]["data"] = fields_array
        else:
            fields_array = utils.split_int_in_bytes(0, length=num_ops * 5)
            for i in range(num_ops):
                start_index = i * 5
                fields_array[start_index] = data[i][0]
                fields_array[start_index + 1: start_index + 5] = utils.split_int_in_bytes(data[i][1], length=4)
            self._packet["payload"]["data"] = fields_array

    def set_read_fields_data(self, data):
        num_ops = len(data)
        if self._packet["header"]["destination"] in [Application.ECG.value, Application.ADPD.value,
                                                     Application.LT_APP.value]:
            fields_array = utils.split_int_in_bytes(0, length=num_ops * 3)
            for i in range(num_ops):
                start_index = i * 3
                fields_array[start_index] = data[i]
            self._packet["payload"]["data"] = fields_array
        else:
            fields_array = utils.split_int_in_bytes(0, length=num_ops * 5)
            for i in range(num_ops):
                start_index = i * 5
                fields_array[start_index] = data[i]
            self._packet["payload"]["data"] = fields_array

    def decode_packet(self, data):
        super().decode_packet(data)
        num_ops = self._packet["payload"]["num_ops"]
        address_value_data = data[11:]
        data = []
        if self._packet["header"]["source"] in [Application.ECG, Application.ADPD, Application.LT_APP]:
            for i in range(num_ops):
                start_index = i * 3
                address_data = address_value_data[start_index:start_index + 1]
                address = utils.join_multi_length_packets(address_data, convert_to_hex=True)
                value_data = address_value_data[start_index + 1:start_index + 3]
                value = utils.join_multi_length_packets(value_data, convert_to_hex=True)
                data.append([address, value])
        else:
            for i in range(num_ops):
                start_index = i * 5
                address_data = address_value_data[start_index:start_index + 1]
                address = utils.join_multi_length_packets(address_data, convert_to_hex=True)
                value_data = address_value_data[start_index + 1:start_index + 5]
                value = utils.join_multi_length_packets(value_data, convert_to_hex=True)
                data.append([address, value])
        self._packet["payload"]["data"] = data


class RegisterPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.ADPD: ['0xC1', '0x10']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x17',
                'checksum': '0x0'
            },
            'payload': {
                'command': <CommonCommand.REGISTER_READ_RES: ['0x22']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'num_ops': 3,
                'data': [
                    [ '0x15', '0x6000' ],
                    [ '0x22', '0x83' ],
                    [ '0x2E', '0x0' ]
                ]
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["num_ops"] = {"size": 1, "join": True}
        self._config["payload"]["data"] = {"size": 2}
        self._packet["payload"]["num_ops"] = [0x01]
        self._packet["payload"]["data"] = [0x00, 0x00, 0x00, 0x00]

    def set_number_of_operations(self, num_ops):
        self._packet["payload"]["num_ops"] = [num_ops]

    def set_data(self, data):
        self._packet["payload"]["data"] = data

    def set_write_reg_data(self, data):
        num_ops = len(data)
        reg_array = utils.split_int_in_bytes(0, length=num_ops * 4)
        for i in range(num_ops):
            start_index = i * 4
            reg_array[start_index:start_index + 2] = utils.split_int_in_bytes(data[i][0], length=2)
            reg_array[start_index + 2:start_index + 4] = utils.split_int_in_bytes(data[i][1], length=2)
        self._packet["payload"]["data"] = reg_array

    def set_read_reg_data(self, data):
        num_ops = len(data)
        reg_array = utils.split_int_in_bytes(0, length=num_ops * 4)
        for i in range(num_ops):
            start_index = i * 4
            reg_array[start_index:start_index + 2] = utils.split_int_in_bytes(data[i], length=2)
        self._packet["payload"]["data"] = reg_array

    def decode_packet(self, data):
        super().decode_packet(data)
        num_ops = self._packet["payload"]["num_ops"]
        address_value_data = data[11:]
        data = []
        for i in range(num_ops):
            start_index = i * 4
            if self._packet["payload"]["command"] == ADPDCommand.CREATE_DCFG_RES:
                address_data = address_value_data[start_index:start_index + 2]
                address = ADPDSlot([utils.join_multi_length_packets(address_data)])
                value_data = address_value_data[start_index + 2:start_index + 4]
                value = ADPDAppID([utils.join_multi_length_packets(value_data)])
            else:
                address_data = address_value_data[start_index:start_index + 2]
                address = utils.join_multi_length_packets(address_data, convert_to_hex=True)
                value_data = address_value_data[start_index + 2:start_index + 4]
                value = utils.join_multi_length_packets(value_data, convert_to_hex=True)
            data.append([address, value])
        self._packet["payload"]["data"] = data


class StreamPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.ADXL: ['0xC1', '0x02']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0xC',
                'checksum': '0x0'
            },
            'payload': {
                'command': <CommonCommand.SUBSCRIBE_STREAM_RES: ['0x0D']>,
                'status': <CommonStatus.SUBSCRIBER_ADDED: ['0x09']>,
                'stream_address': <Stream.ADXL: ['0xC2', '0x02']>
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["stream_address"] = {"size": 2}
        self._packet["payload"]["stream_address"] = [0x00, 0x00]

    def set_stream_address(self, stream_address):
        self._packet["payload"]["stream_address"] = list(reversed(stream_address.value))

    def decode_packet(self, data):
        super().decode_packet(data)
        # if not self._packet["payload"]["stream_address"] == [0, 0]:
        stream_address = self._packet["payload"]["stream_address"]
        self._packet["payload"]["stream_address"] = Stream(list(reversed(stream_address)))


class StreamStatusPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.ADXL: ['0xC1', '0x02']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0xE',
                'checksum': '0x0'
            },
            'payload': {
                'command': <CommonCommand.GET_SENSOR_STATUS_RES: ['0x11']>,
                'status': <CommonStatus.STREAM_STOPPED: ['0x03']>,
                'stream_address': <Stream.ADXL: ['0xC2', '0x02']>,
                'num_subscribers': 0,
                'num_start_registered': 0
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["stream_address"] = {"size": 2}
        self._config["payload"]["num_subscribers"] = {"size": 1, "join": True}
        self._config["payload"]["num_start_registered"] = {"size": 1, "join": True}
        self._packet["payload"]["stream_address"] = [0x00, 0x00]
        self._packet["payload"]["num_subscribers"] = [0x00]
        self._packet["payload"]["num_start_registered"] = [0x00]

    def set_stream_address(self, stream_address):
        self._packet["payload"]["stream_address"] = list(reversed(stream_address.value))

    def decode_packet(self, data):
        super().decode_packet(data)
        if not self._packet["payload"]["stream_address"] == [0, 0]:
            stream_address = list(reversed(self._packet["payload"]["stream_address"]))
            self._packet["payload"]["stream_address"] = Stream(stream_address)


class VersionPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.ADPD: ['0xC1', '0x10']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x42',
                'checksum': '0x0'
            },
            'payload': {
                'command': <CommonCommand.GET_VERSION_RES: ['0x01']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'major_version': 0,
                'minor_version': 3,
                'patch_version': 1,
                'version_string': 'ADPD_App',
                'build_version': 'TEST ADPD4000_VERSION STRING'
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["major_version"] = {"size": 2, "join": True}
        self._config["payload"]["minor_version"] = {"size": 2, "join": True}
        self._config["payload"]["patch_version"] = {"size": 2, "join": True}
        self._config["payload"]["version_string"] = {"size": 10}
        self._config["payload"]["build_version"] = {"size": 40}

    def decode_packet(self, data):
        super().decode_packet(data)
        version_string = ""
        build_version = ""
        for s in self._packet["payload"]["version_string"]:
            if not int(s):
                break
            version_string += chr(s)
        for s in self._packet["payload"]["build_version"]:
            if not int(s):
                break
            build_version += chr(s)
        self._packet["payload"]["version_string"] = str(version_string).encode('utf8').decode("utf-8")
        self._packet["payload"]["build_version"] = build_version


class DFTPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.BCM: ['0xC3', '0x07']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0xC',
                'checksum': '0x0'
            },
            'payload': {
                'command': <BCMCommand.SET_DFT_NUM_RES: ['0x47']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'dft_window': <BCMDFTWindow.DFT_WINDOW_16: ['0x00', '0x02']>
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["dft_window"] = {"size": 2}
        self._packet["payload"]["dft_window"] = [0x00]

    def set_dft_num(self, dft_window):
        self._packet["payload"]["dft_window"] = dft_window.value

    def decode_packet(self, data):
        super().decode_packet(data)
        if self._packet["header"]["source"] == Application.EDA:
            self._packet["payload"]["dft_window"] = EDADFTWindow(self._packet["payload"]["dft_window"])
        else:
            self._packet["payload"]["dft_window"] = BCMDFTWindow(self._packet["payload"]["dft_window"])



# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************

from enum import Enum, unique

from .. import utils


@unique
class ADPDCommand(Enum):
    """
    ADPDCommand Enum
    """
    LOWEST = [0x40]
    LOAD_CONFIG_REQ = [0x42]
    LOAD_CONFIG_RES = [0x43]
    CLOCK_CALIBRATION_REQ = [0x44]
    CLOCK_CALIBRATION_RES = [0x45]
    SET_SLOT_REQ = [0x46]
    SET_SLOT_RES = [0x47]
    GET_SLOT_REQ = [0x48]
    GET_SLOT_RES = [0x49]
    COMMAND_DO_TEST1_REQ = [0x60]
    COMMAND_DO_TEST1_RES = [0x61]
    COMMAND_DO_TEST2_REQ = [0x62]
    COMMAND_DO_TEST2_RES = [0x63]
    COMMAND_DO_TEST3_REQ = [0x64]
    COMMAND_DO_TEST3_RES = [0x65]
    SET_PAUSE_REQ = [0x66]
    SET_PAUSE_RES = [0x67]
    COMMUNICATION_MODE_REQ = [0x68]
    COMMUNICATION_MODE_RES = [0x69]
    SET_SLOT_ACTIVE_REQ = [0x6A]
    SET_SLOT_ACTIVE_RES = [0x6B]
    GET_SLOT_ACTIVE_REQ = [0x6C]
    GET_SLOT_ACTIVE_RES = [0x6D]
    CREATE_DCFG_REQ = [0x70]
    CREATE_DCFG_RES = [0x71]
    SET_SAMPLING_FREQUENCY_REQ = [0x72]
    SET_SAMPLING_FREQUENCY_RES = [0x73]
    AGC_ON_OFF_REQ = [0x76]
    AGC_ON_OFF_RES = [0x77]
    AGC_INFO_REQ = [0x7A]
    AGC_INFO_RES = [0x7B]
    AGC_STATUS_REQ = [0x7C]
    AGC_STATUS_RES = [0x7D]
    SET_EXT_DATA_STREAM_ODR_REQ = [0x7E]
    SET_EXT_DATA_STREAM_ODR_RES = [0x7F]
    EXT_ADPD_DATA_STREAM = [0x80]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))


class ADPDDevice(Enum):
    """
    ADPDDevice Enum
    """
    DEVICE_GREEN = [0x28]
    DEVICE_RED = [0x29]
    DEVICE_INFRARED = [0x2A]
    DEVICE_BLUE = [0x2B]
    DEVICE_G_R_IR_B = [0x2C]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))


class ADPDLed(Enum):
    """
    ADPDLed Enum
    """
    LED_MWL = [0x0]
    LED_GREEN = [0x1]
    LED_RED = [0x2]
    LED_IR = [0x3]
    LED_BLUE = [0x4]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))


class ADPDSlot(Enum):
    """
    ADPDSlot Enum
    """
    SLOT_A = [0x01]
    SLOT_B = [0x02]
    SLOT_C = [0x03]
    SLOT_D = [0x04]
    SLOT_E = [0x05]
    SLOT_F = [0x06]
    SLOT_G = [0x07]
    SLOT_H = [0x08]
    SLOT_I = [0x09]
    SLOT_J = [0x0A]
    SLOT_K = [0x0B]
    SLOT_L = [0x0C]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))


class Clock(Enum):
    """
    Clock Enum
    """
    NO_CLOCK = [0x00]
    CLOCK_32K = [0x01]
    CLOCK_1M = [0x02]
    CLOCK_32M = [0x04]
    CLOCK_32K_AND_1M = [0x05]
    CLOCK_1M_AND_32M = [0x06]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))


class ADPDAppID(Enum):
    """
    ADPDAppID Enum
    """
    APP_ECG = [0]
    APP_PPG = [1]
    APP_TEMPERATURE_THERMISTOR = [2]
    APP_TEMPERATURE_RESISTOR = [3]
    APP_ADPD_GREEN = [4]
    APP_ADPD_RED = [5]
    APP_ADPD_INFRARED = [6]
    APP_ADPD_BLUE = [7]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))



# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************
import math
import time
import logging
from typing import List, Dict, Callable, Optional

from tqdm import tqdm

from ..core import utils
from .fs_application import FSApplication
from ..core.enums.ppg_enums import PPGCommand
from ..core.enums.bcm_enums import BCMCommand
from ..core.enums.adxl_enums import ADXLCommand
from .common_application import CommonApplication
from ..core.enums.fs_enums import FSCommand, FSStatus
from ..core.packets.ppg_packets import PPGStatesPacket
from ..core.packets.command_packet import CommandPacket
from ..core.packets.display_packets import DisplayPacket
from ..core.enums.adpd_enums import ADPDCommand, ADPDLed
from ..core.packets.adxl_packets import ADXLConfigPacket
from ..core.enums.display_enums import DisplayColor, DisplayCommand
from ..core.packets.adpd_packets import AgcInfoPacket, TestCommandPacket
from ..core.packets.bcm_packets import FdsStatusPacket, DCBTimingInfoPacket
from ..core.enums.pm_enums import PMCommand, PowerMode, LDO, ElectrodeSwitch
from ..core.enums.common_enums import Application, CommonCommand, Stream, CommonStatus
from ..core.packets.pm_packets import LDOControlPacket, SwitchControlPacket, PingPacket
from ..core.packets.fs_packets import PatternWritePacket, DebugInfoPacket, BadBlockPacket
from ..core.packets.fs_packets import StreamDebugInfoPacket, FileInfoPacket, PageInfoPacket
from ..core.packets.stream_data_packets import KeyStreamDataPacket, CapSenseStreamDataPacket
from ..core.packets.pm_packets import AppsHealthPacket, ControlPacket, PowerStatePacket, BatteryThresholdPacket

logger = logging.getLogger(__name__)


class TestApplication(CommonApplication):
    WHITE = DisplayColor.WHITE
    BLACK = DisplayColor.BLACK
    RED = DisplayColor.RED
    GREEN = DisplayColor.GREEN
    BLUE = DisplayColor.BLUE

    POWER_MODE_ACTIVE = PowerMode.ACTIVE
    POWER_MODE_HIBERNATE = PowerMode.HIBERNATE
    POWER_MODE_SHUTDOWN = PowerMode.SHUTDOWN

    LDO_FS = LDO.FS
    LDO_OPTICAL = LDO.OPTICAL
    LDO_EPHYZ = LDO.EPHYZ

    SWITCH_AD8233 = ElectrodeSwitch.AD8233
    SWITCH_AD5940 = ElectrodeSwitch.AD5940
    SWITCH_ADPD4000 = ElectrodeSwitch.ADPD4000

    STREAM_ADPD1 = Stream.ADPD1
    STREAM_ADPD2 = Stream.ADPD2
    STREAM_ADPD3 = Stream.ADPD3
    STREAM_ADPD4 = Stream.ADPD4
    STREAM_ADPD5 = Stream.ADPD5
    STREAM_ADPD6 = Stream.ADPD6
    STREAM_ADPD7 = Stream.ADPD7
    STREAM_ADPD8 = Stream.ADPD8
    STREAM_ADPD9 = Stream.ADPD9
    STREAM_ADPD10 = Stream.ADPD10
    STREAM_ADPD11 = Stream.ADPD11
    STREAM_ADPD12 = Stream.ADPD12
    STREAM_ADXL = Stream.ADXL
    STREAM_BCM = Stream.BCM
    STREAM_ECG = Stream.ECG
    STREAM_EDA = Stream.EDA
    STREAM_PEDOMETER = Stream.PEDOMETER
    STREAM_PPG = Stream.PPG
    STREAM_TEMPERATURE = Stream.TEMPERATURE
    STREAM_SYNC_PPG = Stream.SYNC_PPG
    STREAM_SQI = Stream.SQI

    LED_MWL = ADPDLed.LED_MWL
    LED_GREEN = ADPDLed.LED_GREEN
    LED_RED = ADPDLed.LED_RED
    LED_IR = ADPDLed.LED_IR
    LED_BLUE = ADPDLed.LED_BLUE

    def __init__(self, key_press_callback_function, cap_sense_callback_function, packet_manager):
        super().__init__(Application.FS, packet_manager)
        self._key_press_callback = key_press_callback_function
        self._cap_sense_callback = cap_sense_callback_function

    # PM
    def flash_reset(self) -> Dict:
        """
        Resets device flash.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_test_application()
            x = application.flash_reset()
            print(x["payload"]["status"])
            # PMStatus.OK

        """
        packet = CommandPacket(Application.PM, PMCommand.FLASH_RESET_REQ)
        return self._send_packet(packet, PMCommand.FLASH_RESET_RES)

    def get_supported_power_states(self) -> List[PowerMode]:
        """
        List all supported power states for PM.

        :return: Array of power states enums.
        :rtype: List

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_test_application()
            x = application.get_supported_power_states()
            print(x)
            # [<PowerState.ACTIVE_MODE: ['0x0']>, ... , <PowerState.SHUTDOWN_MODE: ['0x3']>]
        """
        return [self.POWER_MODE_ACTIVE, self.POWER_MODE_HIBERNATE, self.POWER_MODE_SHUTDOWN]

    def _power_mode_helper(self, power_mode: PowerMode) -> PowerMode:
        """
        Confirms power mode is from list of Enums.
        """
        if power_mode in self.get_supported_power_states():
            return power_mode
        else:
            logger.warning(
                f"{power_mode} is not supported power mode, choosing {self.get_supported_power_states()[0]} "
                f"as default power mode. use get_supported_power_states() to know all supported power modes.")
            return self.get_supported_power_states()[0]

    def set_power_mode(self, power_state: PowerMode) -> Dict:
        """
        Set specified power state to PM.

        :param power_state: power state to set, use get_supported_power_states() to list all supported power states.
        :type power_state: PowerMode
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_test_application()
            x = application.get_supported_power_states()
            print(x)
            # [<PowerState.ACTIVE_MODE: ['0x0']>, ... , <PowerState.SHUTDOWN_MODE: ['0x3']>]
            x = application.set_power_mode(application.ACTIVE_MODE)
            print(x["payload"]["power_state"])
            # PowerState.ACTIVE_MODE

        """
        power_state = self._power_mode_helper(power_state)
        packet = PowerStatePacket(Application.PM, PMCommand.SET_POWER_STATE_REQ)
        packet.set_state(power_state)
        return self._send_packet(packet, PMCommand.SET_POWER_STATE_RES)

    def set_battery_threshold(self, low_level: int, critical_level: int) -> Dict:
        """
        Set low and critical level threshold for device battery.

        :param low_level: low level threshold for device battery.
        :type low_level: int
        :param critical_level: critical level threshold for device battery.
        :type critical_level: int
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_test_application()
            x = application.set_battery_threshold(20, 5)
            print(x["payload"]["status"])
            # PMStatus.OK

        """
        low_level = utils.range_and_type_check(low_level, type_of=int, lower_and_upper_bound=[0, 100])
        critical_level = utils.range_and_type_check(critical_level, type_of=int, lower_and_upper_bound=[0, 100])
        packet = BatteryThresholdPacket(Application.PM, PMCommand.SET_BAT_THR_REQ)
        packet.set_low_level([low_level])
        packet.set_critical_level([critical_level])
        return self._send_packet(packet, PMCommand.SET_BAT_THR_RES)

    def get_supported_ldo(self) -> List[LDO]:
        """
        List all supported ldo for PM.

        :return: Array of ldo enums.
        :rtype: List

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_test_application()
            x = application.get_supported_ldo()
            print(x)
            # [<LDO.LDO_FS: ['0x1']>, <LDO.OPTICAL_LDO: ['0x2']>, <LDO.LDO_EPHYZ: ['0x3']>]
        """
        return [self.LDO_FS, self.LDO_OPTICAL, self.LDO_EPHYZ]

    def _ldo_helper(self, ldo_id: LDO) -> LDO:
        """
        Confirms ldo id is from list of Enums.
        """
        if ldo_id in self.get_supported_ldo():
            return ldo_id
        else:
            logger.warning(f"{ldo_id} is not supported ldo id, choosing {self.get_supported_ldo()[0]} "
                           f"as default ldo id. use get_supported_ldo() to know all supported ldo IDs")
            return self.get_supported_ldo()[0]

    def disable_ldo(self, ldo_id: LDO) -> Dict:
        """
        Disables specified ldo ID.

        :param ldo_id: ldo ID to disable, use get_supported_ldo() to list all supported ldo IDs.
        :type ldo_id: LDO
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_test_application()
            x = application.get_supported_ldo()
            print(x)
            # [<LDO.LDO_FS: ['0x1']>, <LDO.OPTICAL_LDO: ['0x2']>, <LDO.LDO_EPHYZ: ['0x3']>]
            x = application.disable_ldo(application.LDO_OPTICAL)
            print(x["payload"]["ldo_name"], x["payload"]["enabled"])
            # LDO.OPTICAL False

        """
        ldo_id = self._ldo_helper(ldo_id)
        packet = LDOControlPacket(Application.PM, PMCommand.LDO_CONTROL_REQ)
        packet.set_name(ldo_id)
        packet.set_enable([0x0])
        return self._send_packet(packet, PMCommand.LDO_CONTROL_RES)

    def enable_ldo(self, ldo_id: LDO) -> Dict:
        """
        Enables specified ldo ID.

        :param ldo_id: ldo ID to enable, use get_supported_ldo() to list all supported ldo IDs.
        :type ldo_id: LDO
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_test_application()
            x = application.get_supported_ldo()
            print(x)
            # [<LDO.LDO_FS: ['0x1']>, <LDO.OPTICAL_LDO: ['0x2']>, <LDO.LDO_EPHYZ: ['0x3']>]
            x = application.enable_ldo(application.LDO_OPTICAL)
            print(x["payload"]["ldo_name"], x["payload"]["enabled"])
            # LDO.OPTICAL True

        """
        ldo_id = self._ldo_helper(ldo_id)
        packet = LDOControlPacket(Application.PM, PMCommand.LDO_CONTROL_REQ)
        packet.set_name(ldo_id)
        packet.set_enable([0x1])
        return self._send_packet(packet, PMCommand.LDO_CONTROL_RES)

    def _switch_helper(self, switch_id: ElectrodeSwitch) -> ElectrodeSwitch:
        """
        Confirms switch id is from list of Enums.
        """
        if switch_id in self.get_supported_switches():
            return switch_id
        else:
            logger.warning(f"{switch_id} is not supported switch id, choosing {self.get_supported_switches()[0]} "
                           f"as default switch id. use get_supported_switches() to know all supported switch IDs")
            return self.get_supported_switches()[0]

    def get_supported_switches(self) -> List[ElectrodeSwitch]:
        """
        List all supported switches for PM.

        :return: Array of switches enums.
        :rtype: List

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_test_application()
            x = application.get_supported_switches()
            print(x)
            # [<ElectrodeSwitch.SWITCH_AD8233: ['0x0']>, ... , <ElectrodeSwitch.SWITCH_ADPD4000: ['0x2']>]
        """
        return [self.SWITCH_AD8233, self.SWITCH_AD5940, self.SWITCH_ADPD4000]

    def enable_electrode_switch(self, switch_name: ElectrodeSwitch) -> Dict:
        """
        Enables specified electrode switch.

        :param switch_name: electrode switch to enable, use get_supported_switches() to list all supported switches.
        :type switch_name: ElectrodeSwitch
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_test_application()
            x = application.get_supported_switches()
            print(x)
            # [<ElectrodeSwitch.AD8233_SWITCH: ['0x0']>, ... , <ElectrodeSwitch.ADPD4000_SWITCH: ['0x2']>]
            x = application.enable_electrode_switch(application.SWITCH_AD5940)
            print(x["payload"]["switch_name"], x["payload"]["enabled"])
            # ElectrodeSwitch.AD5940 True

        """
        switch_name = self._switch_helper(switch_name)
        packet = SwitchControlPacket(Application.PM, PMCommand.SW_CONTROL_REQ)
        packet.set_name(switch_name)
        packet.set_enable([0x1])
        return self._send_packet(packet, PMCommand.SW_CONTROL_RES)

    def disable_electrode_switch(self, switch_name: ElectrodeSwitch) -> Dict:
        """
        Disables specified electrode switch.

        :param switch_name: electrode switch to disable, use get_supported_switches() to list all supported switches.
        :type switch_name: ElectrodeSwitch
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_test_application()
            x = application.get_supported_switches()
            print(x)
            # [<ElectrodeSwitch.AD8233_SWITCH: ['0x0']>, ... , <ElectrodeSwitch.ADPD4000_SWITCH: ['0x2']>]
            x = application.disable_electrode_switch(application.SWITCH_AD5940)
            print(x["payload"]["switch_name"], x["payload"]["enabled"])
            # ElectrodeSwitch.AD5940 False

        """
        switch_name = self._switch_helper(switch_name)
        packet = SwitchControlPacket(Application.PM, PMCommand.SW_CONTROL_REQ)
        packet.set_name(switch_name)
        packet.set_enable([0x0])
        return self._send_packet(packet, PMCommand.SW_CONTROL_RES)

    def get_apps_health_status(self) -> Dict:
        """
        Returns ISR count of ad5940, adpd, adxl.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_test_application()
            x = application.get_apps_health_status()
            print(x["payload"]["ad5940_isr_count"], x["payload"]["adpd_isr_count"], x["payload"]["adxl_isr_count"])
            # 0 0 0
        """
        packet = AppsHealthPacket(Application.PM, PMCommand.GET_APPS_HEALTH_REQ)
        return self._send_packet(packet, PMCommand.GET_APPS_HEALTH_RES)

    def disable_cap_sense_test(self) -> Dict:
        """
        Disable cap sense test.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_test_application()
            x = application.disable_cap_sense_test()
            print(x["payload"]["status"])
            # CommonStatus.OK
        """
        packet = ControlPacket(Application.PM, PMCommand.CAP_SENSE_TEST_REQ)
        packet.set_enable([0x0])
        response_packet = self._send_packet(packet, PMCommand.CAP_SENSE_TEST_RES)
        data_packet_id = self._get_packet_id(PMCommand.CAP_SENSE_STREAM_DATA, packet.get_destination())
        self._packet_manager.unsubscribe(data_packet_id, self._cap_sense_callback_data)
        return response_packet

    def enable_cap_sense_test(self) -> Dict:
        """
        Enables cap sense test.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_test_application()
            x = application.enable_cap_sense_test()
            print(x["payload"]["status"])
            # CommonStatus.OK
        """
        packet = ControlPacket(Application.PM, PMCommand.CAP_SENSE_TEST_REQ)
        packet.set_enable([0x1])
        data_packet_id = self._get_packet_id(PMCommand.CAP_SENSE_STREAM_DATA, packet.get_destination())
        self._packet_manager.subscribe(data_packet_id, self._cap_sense_callback_data)
        return self._send_packet(packet, PMCommand.CAP_SENSE_TEST_RES)

    def set_cap_sense_callback(self, callback_function: Callable) -> None:
        """
        Sets the callback for the stream data.

        :param callback_function: callback function for stream key test data.
        :return: None

        .. code-block:: python3
            :emphasize-lines: 7

            from adi_study_watch import SDK

            def callback(data):
                print(data)

            sdk = SDK("COM4")
            application = sdk.get_test_application()
            application.set_cap_sense_callback(callback)
        """
        self._cap_sense_callback = callback_function

    def _cap_sense_callback_data(self, packet, packet_id) -> None:
        """
        Callback for data packets.
        """
        if self._cap_sense_callback:
            try:
                response_packet = CapSenseStreamDataPacket()
                response_packet.decode_packet(packet)
                self._cap_sense_callback(response_packet.get_dict())
            except Exception as e:
                logger.error(f"Can't send packet back to user callback function, reason :: {e}", exc_info=True)
        else:
            logger.warning("No callback function provided")

    def ping(self, num_pings: int, packet_size: int) -> List[Dict]:
        """
        Pings the device to send response of specified packet size and specified times (num_pings).

        :param num_pings: number of times packets to be sent from device.
        :type num_pings: int
        :param packet_size: size of the ping packet from device, must be between 15 and 244.
        :type packet_size: int
        :return: list of response packet as dictionary.
        :rtype: List

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_test_application()
            x = application.ping(3, 16)
            for packet in x:
                print(packet["payload"]["sequence_num"])
            # 1
            # 2
            # 3

        """
        packet_size = utils.range_and_type_check(packet_size, type_of=int, lower_and_upper_bound=[15, 244])
        num_pings = utils.range_and_type_check(num_pings, type_of=int, num_bytes=4)
        packet = PingPacket(Application.PM, CommonCommand.PING_REQ)
        num_pings_arr = utils.split_int_in_bytes(num_pings, length=4)
        packet.set_num_pings(num_pings_arr)
        packet_size = utils.split_int_in_bytes(packet_size, length=2, reverse=True)
        packet.set_packet_size(packet_size)
        packet_id = self._get_packet_id(CommonCommand.PING_RES, packet.get_destination())
        queue = self._get_queue(packet_id)
        self._packet_manager.subscribe(packet_id, self._callback_command)
        self._packet_manager.send_packet(packet)
        result = []
        for _ in range(num_pings):
            start_time = time.time()
            data = self._get_queue_data(queue)
            packet = PingPacket()
            packet.decode_packet(data)
            packet_dict = packet.get_dict()
            elapsed_time = time.time() - start_time
            packet_dict["payload"]["elapsed_time"] = elapsed_time
            result.append(packet_dict)

        self._packet_manager.unsubscribe(packet_id, self._callback_command)
        return result

    # BCM
    def get_fds_status(self) -> Dict:
        """
        FDS status.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_test_application()
            application.get_fds_status()
        """
        packet = FdsStatusPacket(Application.BCM, BCMCommand.FDS_STATUS_REQ)
        return self._send_packet(packet, BCMCommand.FDS_STATUS_RES)

    def read_device_configuration_block_info(self) -> Dict:
        """
        Read Device config block info.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_test_application()
            application.read_device_configuration_block_info()
        """
        packet = DCBTimingInfoPacket(Application.BCM, BCMCommand.DCB_TIMING_INFO_REQ)
        return self._send_packet(packet, BCMCommand.DCB_TIMING_INFO_RES)

    # FS
    def _stream_helper(self, stream: Stream) -> Stream:
        """
        Confirms stream is from list of Enums.
        """
        if stream in self.get_supported_streams():
            return stream
        else:
            logger.warning(f"{stream} is not supported stream, choosing {self.get_supported_streams()[13]} "
                           f"as default stream. use get_supported_streams() to know all supported streams.")
            return self.get_supported_streams()[13]

    def get_supported_streams(self) -> List[Stream]:
        """
        List all supported streams for FS.

        :return: Array of stream ID enums.
        :rtype: List

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_test_application()
            x = application.get_supported_streams()
            print(x)
            # [<Stream.ADPD1: ['0xC2', '0x11']>, ... ]
        """
        return [self.STREAM_ADPD1, self.STREAM_ADPD2, self.STREAM_ADPD3, self.STREAM_ADPD4, self.STREAM_ADPD5,
                self.STREAM_ADPD6, self.STREAM_ADPD7, self.STREAM_ADPD8, self.STREAM_ADPD9, self.STREAM_ADPD10,
                self.STREAM_ADPD11, self.STREAM_ADPD12, self.STREAM_ADXL, self.STREAM_SQI, self.STREAM_ECG,
                self.STREAM_EDA, self.STREAM_PEDOMETER, self.STREAM_PPG, self.STREAM_BCM, self.STREAM_TEMPERATURE,
                self.STREAM_SYNC_PPG]

    def pattern_write(self, file_size: int, scale_type: int, scale_factor: int, base: int,
                      num_files_to_write: int, display_progress=False) -> List[Dict]:
        """
        Pattern Write.
        16384 0 2 1 2 (linear scale)
        16384 1 2 2 2 (log scale)
        16384 2 2 2 2 (exp scale)

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_test_application()
            application.pattern_write(16384, 0, 2, 1, 2)
        """
        result = []
        progress_bar = None
        file_count = 0
        if display_progress:
            progress_bar = tqdm(total=num_files_to_write)
        while file_count < num_files_to_write:
            packet = PatternWritePacket(Application.FS, FSCommand.PATTERN_WRITE_REQ)
            file_size_arr = utils.split_int_in_bytes(file_size, length=4)
            scale_factor_arr = utils.split_int_in_bytes(scale_factor, length=2)
            num_files_to_write_arr = utils.split_int_in_bytes(num_files_to_write, length=2)
            packet.set_pattern_data(file_size_arr, [scale_type], scale_factor_arr, num_files_to_write_arr)
            response_packet = self._send_packet(packet, FSCommand.PATTERN_WRITE_RES)
            result.append(response_packet)
            if response_packet["payload"]["status"] == CommonStatus.OK:
                if scale_type == 0:
                    file_size *= scale_factor
                elif scale_type == 1 and not scale_factor == 1:
                    file_size *= int(math.log(base, scale_factor))
                elif scale_type == 2:
                    file_size *= int(math.exp(scale_factor))
            else:
                if response_packet["payload"]["status"] == FSStatus.ERR_MEMORY_FULL:
                    logger.error("Memory full breaking loop as new files cannot be written!")
                    break
                elif response_packet["payload"]["status"] == FSStatus.ERR_MAX_FILE_COUNT:
                    logger.error("Max file count crossed!")
                    break
            file_count += 1
            if display_progress:
                progress_bar.update(1)
        if display_progress:
            progress_bar.close()
        return result

    def fs_log_test(self) -> Dict:
        """
        Log test.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_test_application()
            application.fs_log_test()
        """
        packet = CommandPacket(Application.FS, FSCommand.TEST_LOG_REQ)
        return self._send_packet(packet, FSCommand.TEST_LOG_RES)

    def get_debug_info(self) -> Dict:
        """
        Returns debug info of FS.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_test_application()
            x = application.get_debug_info()
            print(x["payload"]["status"])
            # CommonStatus.OK
        """
        packet = DebugInfoPacket(Application.FS, FSCommand.GET_DEBUG_INFO_REQ)
        return self._send_packet(packet, FSCommand.GET_DEBUG_INFO_RES)

    def get_bad_blocks(self) -> Dict:
        """
        Returns a packet containing number of bad blocks in the file system.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_test_application()
            x = application.get_bad_blocks()
            print(x["payload"]["bad_blocks"])
            # 0
        """
        packet = BadBlockPacket(Application.FS, FSCommand.GET_BAD_BLOCKS_REQ)
        return self._send_packet(packet, FSCommand.GET_BAD_BLOCKS_RES)

    def get_stream_debug_info(self, stream: Stream) -> Dict:
        """
        Returns specified stream debug info of FS.

        :param stream: stream to obtain debug info, use get_supported_streams() to list all supported streams.
        :type stream: Stream
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_test_application()
            x = application.get_supported_streams()
            print(x)
            # [<Stream.ADPD1: ['0xC2', '0x11']>, ... , <Stream.SQI: ['0xC8', '0x0D']>]
            x = application.get_stream_debug_info(application.STREAM_ADXL)
            print(x["payload"]["stream_address"], x["payload"]["packets_received"], x["payload"]["packets_missed"])
            # Stream.ADXL 0 0
        """
        stream = self._stream_helper(stream)
        packet = StreamDebugInfoPacket(Application.FS, FSCommand.STREAM_DEBUG_INFO_REQ)
        packet.set_stream_address(stream)
        return self._send_packet(packet, FSCommand.STREAM_DEBUG_INFO_RES)

    def get_file_info(self, file_index: int) -> Dict:
        """
        File info.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_test_application()
            x = application.get_file_info(4)

        """
        packet = FileInfoPacket(Application.FS, FSCommand.GET_FILE_INFO_REQ)
        packet.set_file_index(file_index)
        return self._send_packet(packet, FSCommand.GET_FILE_INFO_RES)

    def file_read_test(self, filename: str) -> Optional[dict]:
        """
        File read test.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_test_application()
            x = application.file_read_test("03124671.LOG")

        """
        fs_app = FSApplication(self._packet_manager)
        files = fs_app.ls()
        file_index = 0
        file_found = False
        for file in files:
            if not file_found:
                file_index += 1
            if file["payload"]["filename"] == filename:
                file_found = True
        if not file_found:
            logger.error(f"{filename} is not present on the device, use ls() to list all the files.")
            return None
        response_packet = self.get_file_info(file_index)
        return response_packet

    def page_read_test(self, page_num: int, num_bytes: int) -> Dict:
        """
        Page read test.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_test_application()
            x = application.page_read_test(300, 20)

        """
        utils.range_and_type_check(page_num, type_of=int, num_bytes=4)
        utils.range_and_type_check(num_bytes, type_of=int, num_bytes=1)
        packet = PageInfoPacket(Application.FS, FSCommand.PAGE_READ_TEST_REQ)
        packet.set_page_num(utils.split_int_in_bytes(page_num, length=4))
        packet.set_num_bytes(num_bytes)
        return self._send_packet(packet, FSCommand.PAGE_READ_TEST_RES)

    # Display
    def get_supported_display_colors(self) -> List[DisplayColor]:
        """
        List all supported Display colors for Display Application.

        :return: Array of Display color enums.
        :rtype: List

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_test_application()
            x = application.get_supported_display_colors()
            print(x)
            # [<DisplayColor.WHITE: ['0x0']>, <DisplayColor.BLACK: ['0x1']>, ... , <DisplayColor.BLUE: ['0x4']>]
        """
        return [self.WHITE, self.BLACK, self.RED, self.GREEN, self.BLUE]

    def _color_helper(self, color: DisplayColor) -> DisplayColor:
        """
        Confirms color is from list of Enums.
        """
        if color in self.get_supported_display_colors():
            return color
        else:
            logger.warning(
                f"{color} is not supported display color, choosing {self.get_supported_display_colors()[0]} "
                f"as default display color. use get_supported_display_colors() to know all supported colors.")
            return self.get_supported_display_colors()[0]

    def disable_back_light(self) -> Dict:
        """
        Disables the back light of the watch.

        :return: A response packet as dictionary
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_test_application()
            x = application.disable_back_light()
            print(x["payload"]["status"])
            # CommonStatus.OK
        """
        packet = DisplayPacket(Application.DISPLAY, DisplayCommand.BACKLIGHT_CONTROL_REQ)
        packet.set_data([0x0])
        return self._send_packet(packet, DisplayCommand.BACKLIGHT_CONTROL_RES)

    def disable_key_press_test(self) -> Dict:
        """"
        Disables and unsubscribe to key press test data.

        :return: A response packet as dictionary
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_test_application()
            x = application.enable_key_press_test()
            print(x["payload"]["status"])
            # CommonStatus.OK
        """
        packet = DisplayPacket(Application.DISPLAY, DisplayCommand.KEY_TEST_REQ)
        packet.set_data([0x0])
        response_packet = self._send_packet(packet, DisplayCommand.KEY_TEST_RES)
        data_packet_id = self._get_packet_id(DisplayCommand.KEY_STREAM_DATA, packet.get_destination())
        self._packet_manager.unsubscribe(data_packet_id, self._key_press_callback_data)
        return response_packet

    def enable_key_press_test(self) -> Dict:
        """"
        Enables and subscribe to key press test data.

        :return: A response packet as dictionary
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_test_application()
            x = application.enable_key_press_test()
            print(x["payload"]["status"])
            # CommonStatus.OK
        """
        packet = DisplayPacket(Application.DISPLAY, DisplayCommand.KEY_TEST_REQ)
        packet.set_data([0x1])
        data_packet_id = self._get_packet_id(DisplayCommand.KEY_STREAM_DATA, packet.get_destination())
        self._packet_manager.subscribe(data_packet_id, self._key_press_callback_data)
        return self._send_packet(packet, DisplayCommand.KEY_TEST_RES)

    def enable_back_light(self) -> Dict:
        """"
        Enables the back light of the watch.

        :return: A response packet as dictionary
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_test_application()
            x = application.enable_back_light()
            print(x["payload"]["status"])
            # CommonStatus.OK
        """
        packet = DisplayPacket(Application.DISPLAY, DisplayCommand.BACKLIGHT_CONTROL_REQ)
        packet.set_data([0x1])
        return self._send_packet(packet, DisplayCommand.BACKLIGHT_CONTROL_RES)

    def set_display_color(self, color: DisplayColor) -> Dict:
        """
        Set the specified color to watch screen, to check for pixel damage.

        :param color: color to set on watch screen, use get_supported_display_colors() to list all supported colors.
        :type color: DisplayColor
        :return: A response packet as dictionary
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_test_application()
            x = application.get_supported_display_colors()
            print(x)
            # [<DisplayColor.WHITE: ['0x0']>, <DisplayColor.BLACK: ['0x1']>, ... , <DisplayColor.BLUE: ['0x4']>]
            x = application.set_display_color(application.BLUE)
            print(x["payload"]["color"])
            # DisplayColor.BLUE
        """
        color = self._color_helper(color)
        packet = DisplayPacket(Application.DISPLAY, DisplayCommand.SET_DISPLAY_REQ)
        packet.set_data(color.value)
        return self._send_packet(packet, DisplayCommand.SET_DISPLAY_RES)

    def _key_press_callback_data(self, packet, packet_id) -> None:
        """
        Callback for key test
        """
        if self._key_press_callback:
            try:
                response_packet = KeyStreamDataPacket()
                response_packet.decode_packet(packet)
                self._key_press_callback(response_packet.get_dict())
            except Exception as e:
                logger.error(f"Can't send packet back to user callback function, reason :: {e}", exc_info=True)
        else:
            logger.warning("No callback function provided.")

    def set_key_press_callback(self, callback_function: Callable) -> None:
        """
        Sets the callback for the stream data.

        :param callback_function: callback function for stream key test data.
        :return: None

        .. code-block:: python3
            :emphasize-lines: 7

            from adi_study_watch import SDK

            def callback(data):
                print(data)

            sdk = SDK("COM4")
            application = sdk.get_test_application()
            application.set_callback(callback)

        """
        self._key_press_callback = callback_function

    # ADPD
    def _led_helper(self, led_id: ADPDLed) -> ADPDLed:
        """
        Confirms led ID is from list of Enums.
        """
        if led_id in self.get_supported_led_ids():
            return led_id
        else:
            logger.warning(f"{led_id} is not supported led ID, choosing {self.get_supported_led_ids()[0]} "
                           f"as default led ID. use get_supported_led_ids() to know all supported led IDs.")
            return self.get_supported_led_ids()[0]

    def get_supported_led_ids(self) -> List[ADPDLed]:
        """
        List all supported led IDs for ADPD.

        :return: Array of Led ID enums.
        :rtype: List

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_test_application()
            x = application.get_supported_led_ids()
            print(x)
            # [<ADPDLed.LED_MWL: ['0x0']>, <ADPDLed.LED_GREEN: ['0x1']>, ... , <ADPDLed.LED_BLUE: ['0x4']>]
        """
        return [self.LED_MWL, self.LED_GREEN, self.LED_RED, self.LED_IR, self.LED_BLUE]

    def get_agc_info(self, led_index: ADPDLed) -> Dict:
        """
        Returns AGC info of specified led index.

        :param led_index: led_index to obtain info, use get_supported_led_ids() to list all supported led ID.
        :type led_index: ADPDLed
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_test_application()
            x = application.get_supported_led_ids()
            print(x)
            # [<ADPDLed.LED_GREEN: ['0x0']>, <ADPDLed.LED_RED: ['0x1']>, ... , <ADPDLed.LED_BLUE: ['0x3']>]
            x = application.get_agc_info(application.LED_GREEN)
            print(x["payload"]["led_index"])
            # ADPDLed.LED_GREEN
            print(x["payload"]["ch1"])
            # [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
            print(x["payload"]["ch2"])
            # [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
            print(x["payload"]["dc0_led_current"])
            # 0
            print(x["payload"]["tia_ch1"])
            # 0
            print(x["payload"]["tia_ch2"])
            # 0
        """
        led_index = self._led_helper(led_index)
        packet = AgcInfoPacket(Application.ADPD, ADPDCommand.AGC_INFO_REQ)
        packet.set_led_index(led_index)
        return self._send_packet(packet, ADPDCommand.AGC_INFO_RES)

    def test_command1(self, data: int) -> Dict:
        """
        ADPD test command.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_test_application()
            x = application.test_command1()

        """
        packet = TestCommandPacket(Application.ADPD, ADPDCommand.COMMAND_DO_TEST1_REQ)
        data = utils.split_int_in_bytes(data, length=12)
        packet.set_data(data)
        return self._send_packet(packet, ADPDCommand.COMMAND_DO_TEST1_RES)

    def test_command2(self, data: int) -> Dict:
        """
        ADPD test command.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_test_application()
            x = application.test_command2()

        """
        packet = TestCommandPacket(Application.ADPD, ADPDCommand.COMMAND_DO_TEST2_REQ)
        data = utils.split_int_in_bytes(data, length=12)
        packet.set_data(data)
        return self._send_packet(packet, ADPDCommand.COMMAND_DO_TEST2_RES)

    def test_command3(self, data: int) -> Dict:
        """
        ADPD test command.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_test_application()
            x = application.test_command3()

        """
        packet = TestCommandPacket(Application.ADPD, ADPDCommand.COMMAND_DO_TEST3_REQ)
        data = utils.split_int_in_bytes(data, length=12)
        packet.set_data(data)
        return self._send_packet(packet, ADPDCommand.COMMAND_DO_TEST3_RES)

    # PPG
    def get_ppg_states(self) -> Dict:
        """
        Get PPG states.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_test_application()
            x = application.get_states()
            print(x["payload"]["states"])
            # [0,0,0, ... ,0,0]

        """
        packet = PPGStatesPacket(Application.PPG, PPGCommand.GET_LAST_STATES_REQ)
        return self._send_packet(packet, PPGCommand.GET_LAST_STATES_RES)

    # ADXL
    def adxl_self_test(self) -> Dict:
        """
        ADXL self test.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_test_application()
            x = application.adxl_self_test()

        """
        packet = ADXLConfigPacket(Application.ADXL, ADXLCommand.SELF_TEST_REQ)
        return self._send_packet(packet, ADXLCommand.SELF_TEST_RES)



# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************

from typing import List, Dict, Callable, Tuple

from .csv_logging import CSVLogger
from .common_stream import CommonStream
from ..core.packets.common_packets import DCFGPacket
from ..core.packets.stream_data_packets import TemperatureDataPacket
from ..core.enums.common_enums import Application, Stream, CommonCommand


class TemperatureApplication(CommonStream):
    """
    Temperature Application class.

    .. code-block:: python3
        :emphasize-lines: 4

        from adi_study_watch import SDK

        sdk = SDK("COM4")
        application = sdk.get_temperature_application()

    """

    def __init__(self, callback_function, packet_manager, args):
        super().__init__(Application.TEMPERATURE, Stream.TEMPERATURE, packet_manager, callback_function, args)

    def get_device_configuration(self) -> List[Dict]:
        """
        Returns device configuration data.

        :return: A response packet as dictionary.
        :rtype: List[Dict]

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_temperature_application()
            x = application.get_device_configuration()
            print(x[0]["payload"]["data"])
            # [['0x9', '0x97'], ['0x7', '0x8FFF'], ['0xB', '0x2F6'], ... ]
        """
        packet = DCFGPacket(self._destination, CommonCommand.GET_DCFG_REQ)
        packet_id = self._get_packet_id(CommonCommand.GET_DCFG_RES)
        queue = self._get_queue(packet_id)
        self._packet_manager.subscribe(packet_id, self._callback_command)
        self._packet_manager.send_packet(packet)
        result = []
        count = 1
        while True:
            data = self._get_queue_data(queue)
            packet = DCFGPacket()
            packet.decode_packet(data)
            packet_dict = packet.get_dict()
            result.append(packet_dict)
            if packet_dict["payload"]["num_tx_packets"] == count or packet_dict["payload"]["num_tx_packets"] == 0:
                break
            count += 1
        self._packet_manager.unsubscribe(packet_id, self._callback_command)
        return result

    def get_sensor_status(self) -> Dict:
        """
        Returns packet with number of subscribers and number of sensor start request registered.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_temperature_application()
            x = application.get_sensor_status()
            print(x["payload"]["num_subscribers"], x["payload"]["num_start_registered"])
            # 0 0

        """
        return super().get_sensor_status()

    def set_callback(self, callback_function: Callable, args: Tuple = ()) -> None:
        """
        Sets the callback for the stream data.

        :param args: optional arguments that will be passed with the callback.
        :param callback_function: callback function for stream Temperature data.
        :return: None

        .. code-block:: python3
            :emphasize-lines: 4,12

            from adi_study_watch import SDK

            # make sure optional arguments have default value to prevent them causing Exceptions.
            def callback(data, optional1=None, optional2=None):
                print(data)

            sdk = SDK("COM4")
            application = sdk.get_temperature_application()
            # these optional arguments can be used to pass file, matplotlib or other objects to manipulate data.
            optional_arg1 = "1"
            optional_arg2 = "2"
            application.set_callback(callback, args=(optional_arg1, optional_arg2))

        """
        super().set_callback(callback_function, args)

    def set_timeout(self, timeout_value: float) -> None:
        """
        Sets the time out for queue to wait for command packet response.

        :param timeout_value: queue timeout value (in sec).
        :type timeout_value: int
        :return: None

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_temperature_application()
            application.set_timeout(10)

        """
        super().set_timeout(timeout_value)

    def start_and_subscribe_stream(self) -> Tuple[Dict, Dict]:
        """
        Starts Temperature sensor and also subscribe to the Temperature stream.

        :return: A response packet as dictionary.
        :rtype: Tuple[Dict, Dict]

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_temperature_application()
            start_sensor, subs_stream = application.start_and_subscribe_stream()
            print(start_sensor["payload"]["status"], subs_stream["payload"]["status"])
            # CommonStatus.STREAM_STARTED CommonStatus.SUBSCRIBER_ADDED
        """
        return super().start_and_subscribe_stream()

    def start_sensor(self) -> Dict:
        """
        Starts Temperature sensor.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_temperature_application()
            start_sensor = application.start_sensor()
            print(start_sensor["payload"]["status"])
            # CommonStatus.STREAM_STARTED
        """
        return super().start_sensor()

    def stop_and_unsubscribe_stream(self) -> Tuple[Dict, Dict]:
        """
        Stops Temperature sensor and also Unsubscribe the Temperature stream.

        :return: A response packet as dictionary.
        :rtype: Tuple[Dict, Dict]

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_temperature_application()
            stop_sensor, unsubscribe_stream = application.stop_and_unsubscribe_stream()
            print(stop_sensor["payload"]["status"], unsubscribe_stream["payload"]["status"])
            # CommonStatus.STREAM_STOPPED CommonStatus.SUBSCRIBER_REMOVED
        """
        return super().stop_and_unsubscribe_stream()

    def stop_sensor(self) -> Dict:
        """
        Stops Temperature sensor.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_temperature_application()
            stop_sensor = application.stop_sensor()
            print(stop_sensor["payload"]["status"])
            # CommonStatus.STREAM_STOPPED
        """
        return super().stop_sensor()

    def subscribe_stream(self) -> Dict:
        """
        Subscribe to the Temperature stream.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_temperature_application()
            subs_stream = application.subscribe_stream()
            print(subs_stream["payload"]["status"])
            # CommonStatus.SUBSCRIBER_ADDED
        """
        return super().subscribe_stream()

    def unsubscribe_stream(self) -> Dict:
        """
        Unsubscribe the Temperature stream.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_temperature_application()
            unsubscribe_stream = application.unsubscribe_stream()
            print(unsubscribe_stream["payload"]["status"])
            # CommonStatus.SUBSCRIBER_REMOVED
        """
        return super().unsubscribe_stream()

    def _callback_data(self, packet, packet_id, callback_function=None, args=None):
        """
        Process and returns the data back to user's callback function.
        """
        self._callback_data_helper(packet, packet_id, TemperatureDataPacket(), callback_function, args)

    def enable_csv_logging(self, filename, header=None) -> None:
        """
        Start logging stream data into CSV.

        :param filename: Name of the CSV file.
        :param header: Header list of the CSV file.
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_temperature_application()
            x = application.enable_csv_logging("temp.csv")
        """
        if header is None:
            header = ["Timestamp", "Skin Temperature", "Impedance"]
        self._csv_logger[Stream.TEMPERATURE] = CSVLogger(filename, header)

    def disable_csv_logging(self) -> None:
        """
        Stops logging stream data into CSV.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_temperature_application()
            x = application.disable_csv_logging()
        """
        if self._csv_logger[Stream.TEMPERATURE]:
            self._csv_logger[Stream.TEMPERATURE].stop_logging()
        self._csv_logger[Stream.TEMPERATURE] = None



# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************

from enum import unique, Enum

from .. import utils


@unique
class PMCommand(Enum):
    """
    PMCommand Enum
    """
    LOWEST = [0x40]
    SET_DATE_TIME_REQ = [0x42]
    SET_DATE_TIME_RES = [0x43]
    GET_BAT_INFO_REQ = [0x44]
    GET_BAT_INFO_RES = [0x45]
    SET_BAT_THR_REQ = [0x46]
    SET_BAT_THR_RES = [0x47]
    SET_POWER_STATE_REQ = [0x48]
    SET_POWER_STATE_RES = [0x49]
    SYS_INFO_REQ = [0x4A]
    SYS_INFO_RES = [0x4B]
    GET_DATE_TIME_REQ = [0x52]
    GET_DATE_TIME_RES = [0x53]
    GET_MCU_VERSION_REQ = [0x58]
    GET_MCU_VERSION_RES = [0x59]
    GET_LOW_TOUCH_LOGGING_STATUS_REQ = [0x66]
    GET_LOW_TOUCH_LOGGING_STATUS_RES = [0x67]
    SYSTEM_RESET_REQ = [0x76]
    SYSTEM_RESET_RES = [0x77]
    SW_CONTROL_REQ = [0x78]
    SW_CONTROL_RES = [0x79]
    LDO_CONTROL_REQ = [0x7A]
    LDO_CONTROL_RES = [0x7B]
    CHIP_ID_REQ = [0x7C]
    CHIP_ID_RES = [0x7D]
    CAP_SENSE_TEST_REQ = [0x7E]
    CAP_SENSE_TEST_RES = [0x7F]
    ENTER_BOOTLOADER_REQ = [0x80]
    ENTER_BOOTLOADER_RES = [0x81]
    CAP_SENSE_STREAM_DATA = [0x82]
    ACTIVATE_TOUCH_SENSOR_REQ = [0x86]
    ACTIVATE_TOUCH_SENSOR_RES = [0x87]
    DEACTIVATE_TOUCH_SENSOR_REQ = [0x88]
    DEACTIVATE_TOUCH_SENSOR_RES = [0x89]
    FLASH_RESET_REQ = [0x8A]
    FLASH_RESET_RES = [0x8B]
    SYSTEM_HW_RESET_REQ = [0x8C]
    SYSTEM_HW_RESET_RES = [0x8D]
    GET_APPS_HEALTH_REQ = [0x90]
    GET_APPS_HEALTH_RES = [0x91]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))


class PMStatus(Enum):
    """
    PMStatus Enum
    """
    FIX_STATUS = [0x0]
    LOWEST = [0x40]
    OK = [0x41]
    ERROR_ARGS = [0x42]
    LOW_TOUCH_LOGGING_ALREADY_STARTED = [0x43]
    CONFIG_FILE_NOT_FOUND = [0x44]
    CONFIG_FILE_READ_ERROR = [0x45]
    ENABLE_USER_CONFIG_LOG_FAILED = [0x46]
    USER_CONFIG_LOG_ENABLED = [0x47]
    DISABLE_USER_CONFIG_LOG_FAILED = [0x48]
    USER_CONFIG_LOG_DISABLED = [0x49]
    LOG_STOPPED_THROUGH_BUTTON_A = [0x4A]
    LOW_TOUCH_LOGGING_IN_PROGRESS = [0x4B]
    LOW_TOUCH_LOGGING_NOT_STARTED = [0x4C]
    LOW_TOUCH_MAX_FILE_ERROR = [0x4D]
    LOW_TOUCH_MEMORY_FULL_ERROR = [0x4E]
    ERROR_RESET = [0x4F]
    ENABLE_DCB_CONFIG_LOG_FAILED = [0x50]
    DCB_CONFIG_LOG_ENABLED = [0x51]
    DISABLE_DCB_CONFIG_LOG_FAILED = [0x52]
    DCB_CONFIG_LOG_DISABLED = [0x53]
    ERROR_NOT_CHKD = [0xFF]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))


class MCUType(Enum):
    """
    MCUType Enum
    """
    MCU_INVALID = [0x0]
    MCU_M3 = [0x1]
    MCU_M4 = [0x2]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))


class ElectrodeSwitch(Enum):
    """
    ElectrodeSwitch Enum
    """
    AD8233 = [0x0]
    AD5940 = [0x1]
    ADPD4000 = [0x2]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))


class LDO(Enum):
    """
    LDO Enum
    """
    FS = [0x1]
    OPTICAL = [0x2]
    EPHYZ = [0x3]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))


class ChipID(Enum):
    """
    ChipID Enum
    """
    ADXL362 = [0x1]
    ADPD4K = [0x2]
    ADP5360 = [0x3]
    AD5940 = [0x4]
    NAND_FLASH = [0x5]
    AD7156 = [0x6]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))


class BatteryStatus(Enum):
    """
    BatteryStatus Enum
    """
    NOT_AVAIL = [0x0]
    NOT_CHARGING = [0x1]
    CHARGING = [0x2]
    COMPLETE = [0x3]
    CHARGE_LDO_MODE = [0x4]
    CHARGE_TIMER_EXPIRED = [0x5]
    DETECTION = [0x6]
    CHARGE_ERROR = [0x7]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))


class PowerMode(Enum):
    """
    PowerMode Enum
    """
    ACTIVE = [0x0]
    HIBERNATE = [0x2]
    SHUTDOWN = [0x3]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))



# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************

from .. import utils
from ..enums.common_enums import Application, Command, Status


class CommandPacket:
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.BCM: ['0xC3', '0x07']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0xC',
                'checksum': '0x0'
            },
            'payload': {
                'command': <BCMCommand.SET_HS_TRANS_IMPEDANCE_AMPLIFIER_CAL_RES: ['0x49']>,
                'status': <CommonStatus.OK: ['0x00']>,
            }
        }
    """

    def __init__(self, destination=None, command=None):
        self._config = dict(header={}, payload={})
        self._config["header"]["source"] = {"size": 2}
        self._config["header"]["destination"] = {"size": 2}
        self._config["header"]["length"] = {"size": 2, "join": True, "reverse": True, "hex": True}
        self._config["header"]["checksum"] = {"size": 2, "join": True, "reverse": True, "hex": True}
        self._config["payload"]["command"] = {"size": 1}
        self._config["payload"]["status"] = {"size": 1}

        self._packet = dict(header={}, payload={})
        self._packet["header"]["source"] = [0x00, 0x00]
        self._packet["header"]["destination"] = destination.value if destination else [0x00, 0x00]
        self._packet["header"]["length"] = [0x00, 0x0c]
        self._packet["header"]["checksum"] = [0x00, 0x00]
        self._packet["payload"]["command"] = command.value if command else [0x00]
        self._packet["payload"]["status"] = [0x00]

        self._request_command = command
        self._destination = destination

    def __str__(self):
        return str(self.get_dict())

    def get_destination(self):
        return self._destination

    def clear_packet(self):
        self._packet = dict(header={}, payload={})

    @staticmethod
    def decode(data, config, start_index):
        data_slice = data[start_index:start_index + config["size"] if not config["size"] == -1 else None]
        # data_slice = data[config["size"][0]:config["size"][1]]
        join = config.get("join", False)
        sign = config.get("signed", False)
        reverse = config.get("reverse", False)
        to_hex = config.get("hex", False)
        if join:
            return utils.join_multi_length_packets(data_slice, sign=sign, reverse=reverse, convert_to_hex=to_hex)
        else:
            return data_slice

    def decode_packet(self, data):
        command = data[8:9]
        source = data[0:2]
        command = Command(command, source)
        start_index = 0
        for key in self._config["header"]:
            self._packet["header"][key] = self.decode(data, self._config["header"][key], start_index)
            if key == "source":
                self._packet["header"][key] = command.get_source()
            elif key == "destination":
                self._packet["header"][key] = Application(self._packet["header"][key])
            start_index += self._config["header"][key]["size"]
        # self._packet["header"]["time"] = datetime.now()

        for key in self._config["payload"]:
            self._packet["payload"][key] = self.decode(data, self._config["payload"][key], start_index)
            if key == "command":
                self._packet["payload"][key] = command.get_enum()
            elif key == "status":
                self._packet["payload"][key] = Status(self._packet["payload"][key], source,
                                                      command.get_enum().value).get_enum()
            start_index += self._config["payload"][key]["size"]

    def get_id(self, from_device=False):
        subscribe_id = []
        if from_device:
            subscribe_id += self._packet["header"]["source"].value
            subscribe_id += self._packet["payload"]["command"].value

        return utils.join_multi_length_packets(subscribe_id)

    def get_dict(self):
        return self._packet

    def get_request_command(self):
        return self._request_command

    def set_source(self, source):
        self._packet["header"]["source"] = source.value

    def to_list(self):
        packet_list = []
        for key in self._packet["header"]:
            packet_list += self._packet["header"][key]
        for key in self._packet["payload"]:
            packet_list += self._packet["payload"][key]
        return packet_list

    def update_packet_length(self):
        length = 0
        for key in self._packet["header"]:
            length += len(self._packet["header"][key])
        for key in self._packet["payload"]:
            length += len(self._packet["payload"][key])

        length = utils.split_int_in_bytes(length, length=2)
        self._packet["header"]["length"] = list(reversed(length))
